from __future__ import division

import hashlib
import logging
import os
import uuid

from six.moves.urllib.parse import quote
from six.moves.xmlrpc_client import ServerProxy

from ..bencode import bencode
from ..scgitransport import SCGITransport

logger = logging.getLogger(__name__)

def create_proxy(url):
    proto = url.split(':')[0].lower()
    if proto == 'scgi':
        url = ':'.join(['http'] + url.split(':')[1:])
        logger.debug('Creating SCGI XMLRPC Proxy with url %r' % url)
        return ServerProxy(url, transport=SCGITransport())
    else:
        logger.debug('Creating Normal XMLRPC Proxy with url %r' % url)
        return ServerProxy(url)

def bitfield_to_string(bitfield):
    """
    Converts a list of booleans into a bitfield
    """
    retval = bytearray((len(bitfield) + 7) // 8)
    
    for piece, bit in enumerate(bitfield):
        if bit:
            retval[piece//8] |= 1 << (7 - piece % 8)
    
    return bytes(retval)
    
class RTorrentClient(object):
    def __init__(self, url, label):
        """
        Initializes a new rtorrent client proxy.
        
        url - The url where rtorrent xmlrpc can be reached. Can be both scgi and http.
        label - The label shown in interfaces like rutorrent.
        """
        self.proxy = create_proxy(url)
        self.label = label
    
    def test_connection(self):
        """
        Tests the XMLRPC proxy, returns tuple with cwd and pid if found.
        """
        methods = self.proxy.system.listMethods()
        assert 'view.list' in methods
        return 'cwd:%r, pid:%r' % (self.proxy.system.cwd(), self.proxy.system.pid())
    
    def get_torrents(self):
        """
        Returns a set of info hashes currently added to the client.
        """
        logger.info('Getting a list of torrent hashes')
        return set(x.lower() for x in self.proxy.download_list())
    
    def _get_mtime(self, path):
        return int(os.stat(path).st_mtime)
    
    def add_torrent(self, torrent, destination_path, files):
        """
        Add a new torrent to rtorrent.
        
        torrent is the decoded file as a python object.
        destination_path is where the links are. The complete files must be linked already.
        files is a list of files found in the torrent.
        """
        destination_path = os.path.abspath(destination_path)
        name = torrent[b'info'][b'name']
        logger.info('Trying to add a new torrent to rtorrent: %r' % name)
        
        psize = torrent[b'info'][b'piece length']
        pieces = len(torrent[b'info'][b'pieces']) // 20
        bitfield = [True] * pieces
        
        torrent[b'libtorrent_resume'] = {b'files': []}
        
        current_position = 0
        for f in files:
            logger.debug('Handling file %r' % f)
            
            result = {b'priority': 0, b'completed': int(f['completed'])}
            if f['completed']:
                result[b'mtime'] = self._get_mtime(os.path.join(destination_path, *f['path']))
            torrent[b'libtorrent_resume'][b'files'].append(result)
            
            last_position = current_position + f['length']
            
            first_piece = current_position // psize
            last_piece = (last_position+psize-1) // psize
            
            for piece in range(first_piece, last_piece):
                logger.debug('Setting piece %s to %s' % (piece, f['completed']))
                bitfield[piece] *= f['completed']
            
            current_position = last_position
        
        if all(bitfield):
            logger.info('This torrent is complete, setting bitfield to chunk count')
            torrent[b'libtorrent_resume'][b'bitfield'] = pieces # rtorrent wants the number of pieces when torrent is complete
        else:
            logger.info('This torrent is incomplete, setting bitfield')
            torrent[b'libtorrent_resume'][b'bitfield'] = bitfield_to_string(bitfield)
        
        torrent_file = os.path.join(destination_path, '__tmp_torrent%s.torrent' % uuid.uuid4())
        with open(torrent_file, 'wb') as f:
            f.write(bencode(torrent))
        
        infohash = hashlib.sha1(bencode(torrent[b'info'])).hexdigest()
        
        cmd = [torrent_file, 'd.set_directory_base="%s"' % os.path.abspath(destination_path)]
        cmd.append('d.set_custom1=%s' % quote(self.label))
        
        logger.info('Sending to rtorrent: %r' % cmd)
        self.proxy.load_start(*cmd)
        os.remove(torrent_file)
        
        return infohash in self.get_torrents()



from __future__ import division, unicode_literals

import os
import hashlib
import logging

from collections import defaultdict

from .bencode import bencode, bdecode
from .humanize import humanize_bytes
from .utils import is_scene_modeable, get_actual_scene_name

logger = logging.getLogger('autotorrent')

class Color:
    BLACK = '\033[90m'
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    PINK = '\033[95m'
    CYAN = '\033[96m'
    WHITE = '\033[97m'
    ENDC = '\033[0m'

COLOR_OK = Color.GREEN
COLOR_MISSING_FILES = Color.RED
COLOR_ALREADY_SEEDING = Color.BLUE
COLOR_FOLDER_EXIST_NOT_SEEDING = Color.YELLOW
COLOR_FAILED_TO_ADD_TO_CLIENT = Color.PINK

class Status:
    OK = 0
    MISSING_FILES = 1
    ALREADY_SEEDING = 2
    FOLDER_EXIST_NOT_SEEDING = 3
    FAILED_TO_ADD_TO_CLIENT = 4

status_messages = {
  Status.OK: '%sOK%s' % (COLOR_OK, Color.ENDC),
  Status.MISSING_FILES: '%sMissing%s' % (COLOR_MISSING_FILES, Color.ENDC),
  Status.ALREADY_SEEDING: '%sSeeded%s' % (COLOR_ALREADY_SEEDING, Color.ENDC),
  Status.FOLDER_EXIST_NOT_SEEDING: '%sExists%s' % (COLOR_FOLDER_EXIST_NOT_SEEDING, Color.ENDC),
  Status.FAILED_TO_ADD_TO_CLIENT: '%sFailed%s' % (COLOR_FAILED_TO_ADD_TO_CLIENT, Color.ENDC),
}

class UnknownLinkTypeException(Exception):
    pass

class AutoTorrent(object):
    def __init__(self, db, client, store_path, add_limit_size, add_limit_percent, delete_torrents, link_type='soft'):
        self.db = db
        self.client = client
        self.store_path = store_path
        self.add_limit_size = add_limit_size
        self.add_limit_percent = add_limit_percent
        self.delete_torrents = delete_torrents
        self.link_type = link_type
        self.torrents_seeded = set()

    def populate_torrents_seeded(self):
        """
        Fetches a list of currently-seeded info hashes
        """
        self.torrents_seeded = set(x.lower() for x in self.client.get_torrents())

    def get_info_hash(self, torrent):
        """
        Creates the info hash of a torrent
        """
        return hashlib.sha1(bencode(torrent[b'info'])).hexdigest()

    def index_torrent(self, torrent):
        """
        Indexes the files in the torrent.
        """
        result = []
        if b'files' in torrent[b'info']: # multifile torrent
            files_sorted = {}
            files = {}
            if b'files' in torrent[b'info']:
                torrent_name = torrent[b'info'][b'name'].decode('utf-8')
                logger.info('Found name %r for torrent' % torrent_name)
                
                i = 0
                path_files = defaultdict(list)
                for f in torrent[b'info'][b'files']:
                    orig_path = [x.decode('utf-8') for x in f[b'path']]
                    path = [torrent_name] + orig_path
                    name = path.pop()
                    
                    path_files[os.path.join(*path)].append({
                        'path': orig_path,
                        'length': f[b'length'],
                    })
                    
                    files_sorted['/'.join(orig_path)] = i
                    i += 1
            
            if self.db.scene_mode:
                scene_paths = set()
                for path, files in path_files.items():
                    if is_scene_modeable(f['path'][-1] for f in files):
                        path = path.split(os.sep)
                        name = get_actual_scene_name(path)
                        if not name:
                            continue
                        
                        while path[-1] != name:
                            path.pop()
                        scene_paths.add(os.path.join(*path))
            
            for path, files in path_files.items():
                if self.db.scene_mode:
                    path = path.split(os.sep)
                    while path and os.path.join(*path) not in scene_paths:
                        path.pop()
                else:
                    path = None
                
                if path:
                    name = path[-1]
                    for f in files:
                        actual_path = self.db.find_scene_file_path(name, f['path'][-1], f['length'])
                        f['actual_path'] = actual_path
                        f['completed'] = actual_path is not None
                    result += files
                else:
                    for f in files:
                        actual_path = self.db.find_file_path(f['path'][-1], f['length'])
                        f['actual_path'] = actual_path
                        f['completed'] = actual_path is not None
                    result += files
            # resort the torrent to fit original ordering
            result = sorted(result, key=lambda x:files_sorted['/'.join(x['path'])])
            
        else: # singlefile torrent
            path = torrent[b'info'][b'name'].decode('utf-8')
            length = torrent[b'info'][b'length']
            actual_path = self.db.find_file_path(path, length)
            
            result.append({
                'actual_path': actual_path,
                'length': length,
                'path': [path],
                'completed': actual_path is not None,
            })
        return result

    def parse_torrent(self, torrent):
        """
        Parses the torrent and finds the physical location of files
        in the torrent
        """
        files = self.index_torrent(torrent)

        found_size, missing_size = 0, 0
        for f in files:
            if f['completed']:
                found_size += f['length']
            else:
                missing_size += f['length']

        return found_size, missing_size, files

    def link_files(self, destination_path, files):
        """
        Links the files to the destination_path if they are found.
        """
        if not os.path.isdir(destination_path):
            os.makedirs(destination_path)
        
        for f in files:
            if f['completed']:
                destination = os.path.join(destination_path, *f['path'])
                
                file_path = os.path.dirname(destination)
                if not os.path.isdir(file_path):
                    logger.debug('Folder %r does not exist, creating' % file_path)
                    os.makedirs(file_path)
    
                logger.debug('Making %s link from %r to %r' % (self.link_type, f['actual_path'], destination))
                
                if self.link_type == 'soft':
                    os.symlink(f['actual_path'], destination)
                elif self.link_type == 'hard':
                    os.link(f['actual_path'], destination)
                else:
                    raise UnknownLinkTypeException('%r is not a known link type' % self.link_type)
    
    def handle_torrentfile(self, path):
        """
        Checks a torrentfile for files to seed, groups them by found / not found.
        The result will also include the total size of missing / not missing files.
        """
        logger.info('Handling file %s' % path)

        torrent = self.open_torrentfile(path)

        if self.check_torrent_in_client(torrent):
            self.print_status(Status.ALREADY_SEEDING, path, 'Already seeded')
            if self.delete_torrents:
                logger.info('Removing torrent %r' % path)
                os.remove(path)
            return Status.ALREADY_SEEDING

        found_size, missing_size, files = self.parse_torrent(torrent)
        missing_percent = (missing_size / (found_size + missing_size)) * 100
        found_percent = 100 - missing_percent
        
        if missing_size and missing_percent > self.add_limit_percent or missing_size > self.add_limit_size:
            logger.info('Files missing from %s, only %3.2f%% found (%s missing)' % (path, found_percent, humanize_bytes(missing_size)))
            self.print_status(Status.MISSING_FILES, path, 'Missing files, only %3.2f%% found (%s missing)' % (found_percent, humanize_bytes(missing_size)))
            return Status.MISSING_FILES

        destination_path = os.path.join(self.store_path, os.path.splitext(os.path.split(path)[1])[0])
        
        if os.path.isdir(destination_path):
            logger.info('Folder exist but torrent is not seeded %s' % destination_path)
            self.print_status(Status.FOLDER_EXIST_NOT_SEEDING, path, 'The folder exist, but is not seeded by torrentclient')
            return Status.FOLDER_EXIST_NOT_SEEDING

        self.link_files(destination_path, files)

        if self.delete_torrents:
            logger.info('Removing torrent %r' % path)
            os.remove(path)
        
        if self.client.add_torrent(torrent, destination_path, files):
            self.print_status(Status.OK, path, 'Torrent added successfully')
            return Status.OK
        else:
            self.print_status(Status.FAILED_TO_ADD_TO_CLIENT, path, 'Failed to send torrent to client')
            return Status.FAILED_TO_ADD_TO_CLIENT
    
    def check_torrent_in_client(self, torrent):
        """
        Checks if a torrent is currently seeded
        """
        info_hash = self.get_info_hash(torrent)
        return info_hash in self.torrents_seeded

    def open_torrentfile(self, path):
        """
        Opens and parses a torrent file
        """
        with open(path, 'rb') as f:
            return bdecode(f.read())

    def print_status(self, status, torrentfile, message):
        print(' %-20s %r %s' % ('[%s]' % status_messages[status], os.path.splitext(os.path.basename(torrentfile))[0], message))



import os
import re

__all__ = [
    'is_scene_modeable',
    'get_actual_scene_name',
]

SCENE_FILE_EXTENSIONS = set([
    '.rar',
    '.mp3',
])

def is_scene_modeable(files):
    """
    Checks if a list of files can be considered scene stuff
    """
    extensions = set(os.path.splitext(f)[1].lower() for f in files)
    return '.sfv' in extensions and SCENE_FILE_EXTENSIONS & extensions

def get_actual_scene_name(path):
    """
    Scans a path for the actual scene release name, e.g. skipping cd1 folders.
    
    Returns None if no scene folder could be found
    """
    path = path[::-1]
    for p in path:
        if not p:
            continue
        
        if re.match('^cd[1-9]$', p, re.IGNORECASE):
            continue
        
        if re.match('^samples?$', p, re.IGNORECASE):
            continue
        
        if re.match('^proofs?$', p, re.IGNORECASE):
            continue
        
        if re.match('^(vob)?sub(title)?s?$', p, re.IGNORECASE):
            continue
        
        return p


from __future__ import unicode_literals

import hashlib
import logging
import os
import shelve

from fnmatch import fnmatch

from .utils import is_scene_modeable, get_actual_scene_name

logger = logging.getLogger(__name__)

class Database(object):
    def __init__(self, db_file, paths, ignore_files, scene_mode):
        self.db = shelve.open(db_file)
        self.db_file = db_file
        self.paths = paths
        self.ignore_files = [self.normalize_filename(x) for x in ignore_files]
        self.scene_mode = scene_mode
    
    def truncate(self):
        """
        Truncates the database
        """
        logger.info('Truncated the database')
        self.db.close()
        self.db = shelve.open(self.db_file, flag='n')
    
    def insert_into_database(self, root, f, is_scene):
        """
        Does the actual insertion into the database.
        """
        path = os.path.abspath(os.path.join(root, f))
        size = os.path.getsize(path)
        normalized_filename = self.normalize_filename(f)
        if is_scene:
            normalized_folder = self.normalize_filename(get_actual_scene_name(root.split(os.sep)))
            key = self.keyify(size, normalized_folder, normalized_filename)
        else:
            key = self.keyify(size, normalized_filename)
        
        if key in self.db: # check if same file
            old_inode = os.stat(self.db[key]).st_ino
            new_inode = os.stat(path).st_ino
            if old_inode != new_inode:
                logger.warning('Duplicate key %s and %s' % (path, self.db[key]))

        self.db[key] = path
    
    def rebuild(self):
        """
        Scans the paths for files and rebuilds the database.
        """
        logger.info('Rebuilding database')
        self.truncate()
        
        scene_paths = set()
        if self.scene_mode:
            logger.info('Scene mode enabled, doing a preliminary scan')
            for root_path in self.paths:
                logger.info('Preliminary scanning %s' % root_path)
                for root, dirs, files in os.walk(root_path):
                    if is_scene_modeable(files):
                        sep_root = root.split(os.sep)
                        name = get_actual_scene_name(root.split(os.sep))
                        while sep_root[-1] != name:
                            sep_root.pop()
                        path = os.path.join(*sep_root)
                        logger.debug('Found scene path %r' % path)
                        scene_paths.add(path)
                logger.info('Done preliminary scanning %s' % root_path)
        
        for root_path in self.paths:
            logger.info('Scanning %s' % root_path)
            for root, dirs, files in os.walk(root_path):
                if self.scene_mode:
                    sep_root = root.split(os.sep)
                    while sep_root:
                        if os.path.join(*sep_root) in scene_paths:
                            break
                        sep_root.pop()
                    
                    if sep_root:
                        logger.info('Looks like we found a scene release in %r for rls %r' % (root, name))
                        for f in files:
                            self.insert_into_database(root, f, True)
                        continue
                
                for f in files:
                    normalized_filename = self.normalize_filename(f)
                    
                    do_skip = False
                    for ignore_file in self.ignore_files:
                        if fnmatch(normalized_filename, ignore_file):
                            do_skip = True
                            break
                    if do_skip:
                        continue
                    
                    self.insert_into_database(root, f, False)

                    
            logger.info('Done scanning %s' % root_path)
        self.db.sync()
    
    def find_scene_file_path(self, rls, f, size):
        """
        Looks for a file in the database.
        """
        key = self.keyify(size, self.normalize_filename(rls), self.normalize_filename(f))

        return self.db.get(key, None)
    
    def find_file_path(self, f, size):
        """
        Looks for a file in the database.
        """
        key = self.keyify(size, self.normalize_filename(f))

        return self.db.get(key, None)
    
    def keyify(self, size, *names):
        """
        Turns a name and size into a key that can be stored in the database.
        """
        key = '%s|%s' % (size, '|'.join(names))
        logger.debug('Keyify: %s' % key)
        
        return hashlib.sha256(key.encode('utf-8')).hexdigest()
    
    def normalize_filename(self, filename):
        """
        Normalizes a filename to better detect simlar files.
        """
        return filename.replace(' ', '_').lower()